--[[ 
	Shadowed Unit Frames (Trinkets), Mayen of Mal'Ganis (US) PvP
]]

local Trinkets = select(2, ...)
LibStub("AceTimer-3.0"):Embed(Trinkets)

local SL = ShadowUF.L or ShadowUFLocals
local L = Trinkets.L
local usedTrinkets, frames = {}, {}
ShadowUF.usedTrinkets = usedTrinkets

local frame = CreateFrame("Frame")
Trinkets.frame = frame
frame:SetScript("OnEvent", function(self, event, ...)
	Trinkets[event](Trinkets, ...)
end)

function Trinkets:ResetTrinket(guid)
	usedTrinkets[guid] = nil
	self:UpdateTrinket(guid)
end

function Trinkets:UpdateTrinket(guid)
	for frame in pairs(frames) do
		if( frame.unit and UnitGUID(frame.unit) == guid ) then
			for _, fontString in pairs(frame.fontStrings) do
				if( fontString.PVP_TRINKETS ) then
					fontString:UpdateTags()
				end
			end
		end
	end
end

function Trinkets:PLAYER_LEAVING_WORLD()
	self:CancelAllTimers()
	table.wipe(usedTrinkets)

	for frame in pairs(frames) do
		for _, fontString in pairs(frame.fontStrings) do
			if( fontString.PVP_TRINKETS ) then
				fontString:UpdateTags()
			end
		end
	end
end

-- Check for PvP trinkets being used
function Trinkets:COMBAT_LOG_EVENT_UNFILTERED(timestamp, eventType, sourceGUID, sourceName, sourceFlags, destGUID, destName, destFlags, spellID, spellName, spellSchool)
	if( eventType ~= "SPELL_CAST_SUCCESS" ) then return end
	
	-- Human Racial / PvP Trinket
	if( spellID == 59752 or spellID == 42292 ) then
		usedTrinkets[sourceGUID] = true

		self:ScheduleTimer("ResetTrinket", 120, sourceGUID)
		self:UpdateTrinket(sourceGUID)
	-- Will of the Forsaken
	elseif( spellID == 7744 ) then
		usedTrinkets[sourceGUID] = true

		self:ScheduleTimer("ResetTrinket", 45, sourceGUID)
		self:UpdateTrinket(sourceGUID)
	end
end

-- Setup our custom tag, using the general tags system so if this addon is removed the tags will no longer show up as their name instead of silently breaking
ShadowUF.Tags.defaultTags["pvp:trinket"] = [[
	function(unit, unitOwner, fontString)
		if( ShadowUF.usedTrinkets[UnitGUID(unit)] ) then return end

		-- Use the appropiate icon for the unit
		local trinketIcon
		if( UnitFactionGroup(unit) == "Horde" ) then
			trinketIcon = UnitLevel(unit) == 80 and "Interface\\Icons\\INV_Jewelry_Necklace_38" or "Interface\\Icons\\INV_Jewelry_TrinketPVP_02"
		else
			trinketIcon = UnitLevel(unit) == 80 and "Interface\\Icons\\INV_Jewelry_Necklace_37" or "Interface\\Icons\\INV_Jewelry_TrinketPVP_01"
		end
		
		return string.format("|T%s:0:0:0:0|t", trinketIcon)
	end
]]

ShadowUF.Tags.defaultCategories["pvp:trinket"] = "pvp"
ShadowUF.Tags.defaultEvents["pvp:trinket"] = "PVP_TRINKETS"
ShadowUF.Tags.defaultNames["pvp:trinket"] = L["PVP Trinket status"]
ShadowUF.Tags.defaultHelp["pvp:trinket"] = L["Returns a pvp trinket if the unit has not used their pvp trinket or Will of the Forsaken, once it's used the icon will be hidden until their pvp trinket is usable again."]

ShadowUF.Tags.defaultTags["pvp:trinket:large"] = [[
	function(unit, unitOwner, fontString)
		if( ShadowUF.usedTrinkets[UnitGUID(unit)] ) then return end

		-- Use the appropiate icon for the unit
		local trinketIcon
		if( UnitFactionGroup(unit) == "Horde" ) then
			trinketIcon = UnitLevel(unit) == 80 and "Interface\\Icons\\INV_Jewelry_Necklace_38" or "Interface\\Icons\\INV_Jewelry_TrinketPVP_02"
		else
			trinketIcon = UnitLevel(unit) == 80 and "Interface\\Icons\\INV_Jewelry_Necklace_37" or "Interface\\Icons\\INV_Jewelry_TrinketPVP_01"
		end
		
		return string.format("|T%s:20:20:0:0|t", trinketIcon)
	end
]]

ShadowUF.Tags.defaultCategories["pvp:trinket:large"] = "pvp"
ShadowUF.Tags.defaultEvents["pvp:trinket:large"] = "PVP_TRINKETS"
ShadowUF.Tags.defaultNames["pvp:trinket:large"] = L["PVP Trinket status (Large)"]
ShadowUF.Tags.defaultHelp["pvp:trinket:large"] = L["Same as [pvp:trinket] but a larger version instead of one that is the same dimensions as your text height."]

-- Maintain the general list of tags that care about pvp trinkets
ShadowUF.Tags.customEvents["PVP_TRINKETS"] = Trinkets
function Trinkets:EnableTag(frame)
	frames[frame] = true
	self.frame:RegisterEvent("COMBAT_LOG_EVENT_UNFILTERED")
end

function Trinkets:DisableTag(frame)
	frames[frame] = nil
	
	local active
	for frame in pairs(frames) do
		active = true
		break
	end

	if( not active ) then
		self.frame:UnregisterEvent("COMBAT_LOG_EVENT_UNFILTERED")
	end
end
